import java.util.concurrent.Semaphore;
import javax.persistence.*;

/*
 Optimistic locking failure:

 I am using a Semaphore to force this execution of concurrents
 transactions T1 and T2 that cause parallel updates to A/B but
 it does not throw an OptimisticLockException and roll back. 

           T1                           T2
 ----------------------------------------------------------
 READ(A) version=1 value=0   |
 READ(B) version=1 value=0   |
                             | READ(A)  version=1 value=0
                             | READ(B)  version=1 value=0
                             | B = A + B + 20
                             | WRITE(B) version=2 value=20
 A = A + B + 10              |
 WRITE(A) version=2 value=10 |
 -----------------------------------------------------------
 Results A: version=2 value=10, B: version=2 value=20
*/
@Entity
public class Test {
    @Version long version;
    @Id @GeneratedValue long id;
    int value;

    @Override
    public String toString() {
        return "B" + id + " version="+version+" value="+value;
    }

    static boolean DEBUG = true;

    private static synchronized void debug(Object obj) {
        if (DEBUG) System.out.println(Thread.currentThread()+" "+obj);
    }
 
    public static void main(String... args) throws Exception {
        execute(true);
        execute(false);
    }

    public static void execute(boolean seq) throws Exception {
        final Semaphore sem = new Semaphore(0);
        final LockModeType lock = LockModeType.WRITE;
        System.out.println("Sequential="+seq+" Lock="+lock);

        final EntityManagerFactory emf =
            Persistence.createEntityManagerFactory("objectdb:test.tmp;drop");
        EntityManager em = emf.createEntityManager();
        em.getTransaction().begin();
        A a = new A();
        Test b = new Test();
        em.persist( a );
        em.persist( b );
        em.getTransaction().commit();

        final Long minA = (Long)em.createQuery("SELECT e.id FROM A e")
                          .getResultList().get(0);
        final Long minB = (Long)em.createQuery("SELECT e.id FROM Test e")
                          .getResultList().get(0);

        Thread t1 = new Thread(new Runnable() {
          public void run() {
            EntityManager em = emf.createEntityManager();
            try{
              em.getTransaction().begin();
              A a = em.find(A.class, minA, lock);
              debug( "READ  " + a );
              Test b = em.find(Test.class, minB, lock);
              debug( "READ  " + b );

              // Continue execution transaction t2
              if (DEBUG) sem.release();
              Thread.yield();
              if (DEBUG) try { sem.acquire();
              } catch(InterruptedException iex) { iex.printStackTrace(); }

              a.value += b.value + 10;
              em.getTransaction().commit();
              debug( "WRITE " + a );
              em.close();
            } finally {
              if (em.getTransaction().isActive()) 
                em.getTransaction().rollback();
              if (DEBUG) sem.release();
            }
          }
        });

        Thread t2 = new Thread(new Runnable() {
          public void run() {
            if (DEBUG) try { sem.acquire();
            } catch(InterruptedException iex) { iex.printStackTrace(); }
            EntityManager em = emf.createEntityManager();
            try {
              em.getTransaction().begin();
              A a = em.find(A.class, minA, lock);
              debug( "READ  " + a );
              Test b = em.find(Test.class, minB, lock);
              debug( "READ  " + b );

              b.value += a.value + 20;
              em.getTransaction().commit();
              debug( "WRITE " + b);
              em.close();
            } finally {
              if (em.getTransaction().isActive()) 
                em.getTransaction().rollback();
              if (DEBUG) sem.release();
            }
          }
        });

        // Threads simulating concurrents transactions
        t1.start(); 
        if ( seq ) //Execution sequential <t1, t2>
            t1.join(); //wait thread t1 end
        t2.start();
        // wait threads end
        t2.join();
        t1.join();

        em.refresh(a);
        em.refresh(b);
        System.out.println( a + ", " + b );
        em.close();
        emf.close();
    }


    @Entity
    static class A {
        @Version long version;
        @Id @GeneratedValue long id;
        int value;

        @Override
        public String toString() {
            return "A" + id + " version="+version+" value="+value;
        }
    }
}
